//
//Example of HDR rendering 
//
// Author: Alex V. Boreskoff <alexboreskoff@mtu-net.ru>, <steps3d@narod.ru>
//

#include	"libExt.h"

#ifdef	MACOSX
	#include	<GLUT/glut.h>
#else
	#include	<glut.h>
#endif

#include	<stdio.h>
#include	<stdlib.h>

#include	"libTexture.h"
#include	"Vector3D.h"
#include	"Vector2D.h"
#include	"Vector4D.h"
#include	"Data.h"
#include	"GlslProgram.h"
#include	"FrameBuffer.h"
#include	"boxes.h"

Vector3D	eye   ( 7, 5, 7 );			// camera position
Vector3D	light ( 5, 0, 4 );			// light position
float	 	angle = 0;
Vector3D	rot ( 0, 0, 0 );
int			mouseOldX = 0;
int			mouseOldY = 0;
float		exposure  = 1.0;
unsigned	diffMap;
unsigned	decalMap;
bool		clampingControl = false;
#define	IMAGE_SIZE	512

FrameBuffer	buffer  ( IMAGE_SIZE, IMAGE_SIZE, FrameBuffer :: depth32 );
FrameBuffer	buffer0 ( IMAGE_SIZE / 2, IMAGE_SIZE / 2, 0 );
FrameBuffer	buffer2 ( IMAGE_SIZE / 4, IMAGE_SIZE / 2, 0 );
FrameBuffer	buffer3 ( IMAGE_SIZE / 4, IMAGE_SIZE / 4, 0 );
GlslProgram	blinnProgram;
GlslProgram	toneMapProgram;
GlslProgram	xBlur;
GlslProgram yBlur;
GlslProgram	filterMap;

void reshape ( int w, int h );

/////////////////////////////////////////////////////////////////////////////////

void init ()
{
	glClearColor ( 0, 0, 0, 1 );
	glEnable     ( GL_DEPTH_TEST );
	glDepthFunc  ( GL_LEQUAL );

	glHint ( GL_POLYGON_SMOOTH_HINT,         GL_NICEST );
	glHint ( GL_PERSPECTIVE_CORRECTION_HINT, GL_NICEST );
}

void	drawQuad ( unsigned tex )
{
	glActiveTextureARB ( GL_TEXTURE0_ARB );
	glColor4f     ( 1, 1, 1, 1 );
    glBindTexture ( GL_TEXTURE_2D, tex );
	
	glEnable      ( GL_TEXTURE_2D );
    glBegin       ( GL_QUADS );

        glTexCoord2f ( 0, 0 );
        glVertex2f   ( 0, 0 );

        glTexCoord2f ( 1, 0 );
        glVertex2f   ( 1, 0 );

        glTexCoord2f ( 1, 1 );
        glVertex2f   ( 1, 1 );

        glTexCoord2f ( 0, 1 );
        glVertex2f   ( 0, 1 );

    glEnd   ();
}

void    startOrtho ()
{
    glMatrixMode   ( GL_PROJECTION );                   // select the projection matrix
    glPushMatrix   ();                                  // store the projection matrix
    glLoadIdentity ();                                  // reset the projection matrix
                                                        // set up an ortho screen
    glOrtho        ( 0, 1, 0, 1, -1, 1 );
    glMatrixMode   ( GL_MODELVIEW );                    // select the modelview matrix
    glPushMatrix   ();                                  // store the modelview matrix
    glLoadIdentity ();                                  // reset the modelview matrix

	glDisable   ( GL_DEPTH_TEST );
	glDepthMask ( GL_FALSE );
}

void    endOrtho ()
{
    glMatrixMode ( GL_PROJECTION );                     // select the projection matrix
    glPopMatrix  ();                                    // restore the old projection matrix
    glMatrixMode ( GL_MODELVIEW );                      // select the modelview matrix
    glPopMatrix  ();                                    // restore the old projection matrix

	glEnable    ( GL_DEPTH_TEST );
	glDepthMask ( GL_TRUE );
}

void	blurMap ()
{
	startOrtho ();
	
	buffer0.bind ();
	filterMap.bind ();
	drawQuad ( buffer.getColorBuffer ( 1 ) );
	filterMap.unbind ();
	buffer0.unbind ();
	
	buffer2.bind ();
	xBlur  .bind ();									// perform x-blurring first
	drawQuad ( buffer0.getColorBuffer () );
	xBlur.unbind   ();
	buffer2.unbind ();

	buffer3.bind ();
	yBlur.bind ();
	drawQuad ( buffer2.getColorBuffer () );
	yBlur.unbind ();
	buffer3.unbind ();
														// now in buffer3 we have blurred FP map
	endOrtho ();
}

void display ()
{
	GLenum	buffers [] = { GL_COLOR_ATTACHMENT0_EXT, GL_COLOR_ATTACHMENT1_EXT };
	
												// render for FP FBO
	buffer.bind ();

	glDrawBuffers ( 2, buffers );
	
	reshape ( buffer.getWidth (), buffer.getHeight () );
	
	glClear ( GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT );

												// draw the light
	glMatrixMode ( GL_MODELVIEW );
	glPushMatrix ();

	if ( clampingControl )
		glClampColorARB    ( GL_CLAMP_VERTEX_COLOR_ARB, GL_FALSE );		// disable vertex color clamping
		
	glDisable          ( GL_TEXTURE_2D );
	glTranslatef       ( light.x, light.y, light.z );
	glColor4f          ( 30, 30, 30, 1 );
	glutSolidSphere    ( 0.1f, 15, 15 );
	glPopMatrix        ();

	glMatrixMode   ( GL_MODELVIEW );
	glPushMatrix   ();

    glRotatef    ( rot.x, 1, 0, 0 );
    glRotatef    ( rot.y, 0, 1, 0 );
    glRotatef    ( rot.z, 0, 0, 1 );

	glBindTexture ( GL_TEXTURE_2D, diffMap );
	
	blinnProgram.bind ();
	
	drawBox ( Vector3D ( -1, -1, -2 ), Vector3D ( 1.7, 1.7, 1.7 ), decalMap );
    
	glTranslatef ( 2.5, 2.5, 1 );
	glRotatef    ( 15*angle, 0, 1, 0 );
	
	glutSolidTorus ( 0.7, 1.6, 20, 20 );
	
	glRotatef    ( 10*angle, 0, 0, 1 );
	
	drawBox ( Vector3D ( -2, -2, 1 ), Vector3D ( 1.5, 1.5, 1.5 ), diffMap );
	
	glEnable ( GL_TEXTURE_2D );
    
	blinnProgram.unbind ();
	buffer.unbind       ();

    glPopMatrix     ();
	
												// now process image (blur)
	blurMap ();

	startOrtho ();
	
	glClear ( GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT );

	glActiveTextureARB ( GL_TEXTURE1_ARB );
	glBindTexture      ( GL_TEXTURE_2D, buffer3.getColorBuffer () );

	toneMapProgram.bind  ();
	drawQuad ( buffer.getColorBuffer () );
	toneMapProgram.unbind ();
	
	endOrtho ();
	
	glutSwapBuffers ();
}

void reshape ( int w, int h )
{
   glViewport     ( 0, 0, (GLsizei)w, (GLsizei)h );
   glMatrixMode   ( GL_PROJECTION );
   											// factor all camera ops into projection matrix
   glLoadIdentity ();
   gluPerspective ( 60.0, (GLfloat)w/(GLfloat)h, 1.0, 60.0 );
   gluLookAt      ( eye.x, eye.y, eye.z,	// eye
					0, 0, 0,				// center
					0.0, 0.0, 1.0 );		// up

   glMatrixMode   ( GL_MODELVIEW );
   glLoadIdentity ();
}

void motion ( int x, int y )
{
    rot.y -= ((mouseOldY - y) * 180.0f) / 200.0f;
    rot.z -= ((mouseOldX - x) * 180.0f) / 200.0f;
    rot.x  = 0;

    if ( rot.z > 360 )
		rot.z -= 360;

	if ( rot.z < -360 )
		rot.z += 360;

    if ( rot.y > 360 )
		rot.y -= 360;

	if ( rot.y < -360 )
        rot.y += 360;

    mouseOldX = x;
    mouseOldY = y;

	glutPostRedisplay ();
}

void mouse ( int button, int state, int x, int y )
{
    if ( state == GLUT_DOWN )
    {
        mouseOldX = x;
        mouseOldY = y;
	}
}

void key ( unsigned char key, int x, int y )
{
	if ( key == 27 || key == 'q' || key == 'Q' )	//	quit requested
    	exit ( 0 );
		
	if ( key == '+' )
		exposure += 0.02;
	else
	if ( key == '-' )
		exposure -= 0.02;
}

void	animate ()
{
	angle  = 0.004f * glutGet ( GLUT_ELAPSED_TIME );

	light.x = 2*cos ( angle );
	light.y = 3*sin ( 1.4 * angle );
	light.z = 3 + 0.5 * sin ( angle / 3 );

	blinnProgram.bind ();
	blinnProgram.setUniformVector ( "eyePos",   Vector4D ( eye,   1 ) );
	blinnProgram.setUniformVector ( "lightPos", Vector4D ( light, 1 ) );
	blinnProgram.unbind ();

	toneMapProgram.bind ();
	toneMapProgram.setUniformFloat ( "exposure",     exposure );
	toneMapProgram.setUniformFloat ( "blurStrength", 4.0 );
	toneMapProgram.unbind ();
	
	glutPostRedisplay ();
}

int main ( int argc, char * argv [] )
{
								// initialize glut
	glutInit            ( &argc, argv );
	glutInitDisplayMode ( GLUT_DOUBLE | GLUT_RGB | GLUT_DEPTH );
	glutInitWindowSize  ( IMAGE_SIZE, IMAGE_SIZE );


								// create window
	glutCreateWindow ( "Example of HDR rendering" );

								// register handlers
	glutDisplayFunc  ( display );
	glutReshapeFunc  ( reshape );
	glutKeyboardFunc ( key     );
	glutMouseFunc    ( mouse   );
	glutMotionFunc   ( motion  );
	glutIdleFunc     ( animate );

	init           ();
	initExtensions ();
	
	assertExtensionsSupported ( "GL_ARB_shading_language_100 GL_ARB_shader_objects" );
	
	clampingControl = isExtensionSupported ( "ARB_color_buffer_float" );
	
	if ( !FrameBuffer :: isSupported () )
	{
		printf ( "Fbo not supported\n" );
		
		return 1;
	}
	
	if ( !GlslProgram :: isSupported () )
	{
		printf ( "GLSL not supported\n" );
		
		return 1;
	}
	
    diffMap = createTexture2D ( true,  "../../Textures/oak.bmp" );
    decalMap = createTexture2D ( true,  "../../Textures/16.jpg" );

	if ( !blinnProgram.loadShaders ( "blinn.vsh", "blinn-2.fsh" ) )
	{
		printf ( "Error loading blinn-2 shaders:\n%s\n", blinnProgram.getLog ().c_str () );

		return 3;
	}

    if ( !toneMapProgram.loadShaders ( "toneMap.vsh", "toneMap-2.fsh" ) )
    {
        printf ( "Error loading toneMap shaders:\n%s\n", toneMapProgram.getLog ().c_str () );

        return 3;
    }
	
    if ( !xBlur.loadShaders ( "b-x.vsh", "b-x-2.fsh" ) )
    {
        printf ( "Error loading xBlur shaders:\n%s\n", xBlur.getLog ().c_str () );

        return 3;
    }

    if ( !yBlur.loadShaders ( "b-y.vsh", "b-y-2.fsh" ) )
    {
        printf ( "Error loading yBlur shaders:\n%s\n", yBlur.getLog ().c_str () );

        return 3;
    }
	
    if ( !filterMap.loadShaders ( "filterMap.vsh", "filterMap-2.fsh" ) )
    {
        printf ( "Error loading filterMap shaders:\n%s\n", yBlur.getLog ().c_str () );

        return 3;
    }
	
	xBlur.bind              ();
	xBlur.setTexture        ( "mainTex", 0 );
	xBlur.unbind            ();
	yBlur.bind              ();
	yBlur.setTexture        ( "mainTex", 0 );
	yBlur.unbind            ();
	blinnProgram.bind       ();
	blinnProgram.setTexture ( "diffMap", 0 );
	blinnProgram.unbind     ();
	
	filterMap.bind       ();
	filterMap.setTexture ( "mainMap", 0 );
	filterMap.unbind     ();
	
	toneMapProgram.bind       ();
	toneMapProgram.setTexture ( "mainMap", 0 );
	toneMapProgram.setTexture ( "blurMap", 1 );
	toneMapProgram.unbind     ();

	GLenum	 format        = GL_RGBA8;
	unsigned screenMapMain = buffer.createColorTexture ( GL_RGBA, format, GL_CLAMP, FrameBuffer :: filterLinear );
	unsigned screenMapAdd  = buffer.createColorTexture ( GL_RGBA, format, GL_CLAMP, FrameBuffer :: filterLinear );
	
	buffer.create ();
	buffer.bind   ();
	
	if ( !buffer.attachColorTexture ( GL_TEXTURE_2D, screenMapMain, 0 ) )
		printf ( "buffer error with color attachment\n");

	if ( !buffer.attachColorTexture ( GL_TEXTURE_2D, screenMapAdd, 1 ) )
		printf ( "buffer error with color attachment\n");

	if ( !buffer.isOk () )
		printf ( "Error with framebuffer\n" );
					
	buffer.unbind ();

	unsigned screenMap0 = buffer0.createColorTexture ( GL_RGBA, format, GL_CLAMP, FrameBuffer :: filterLinear );
	
	buffer0.create ();
	buffer0.bind   ();
	if ( !buffer0.attachColorTexture ( GL_TEXTURE_2D, screenMap0 ) )
		printf ( "buffer0 error with color attachment\n");

	if ( !buffer0.isOk () )
		printf ( "Error with framebuffer0\n" );
					
	buffer0.unbind ();

	unsigned screenMap2 = buffer2.createColorTexture ( GL_RGBA, format, GL_CLAMP, FrameBuffer :: filterLinear );

	buffer2.create ();
	buffer2.bind   ();
	if ( !buffer2.attachColorTexture ( GL_TEXTURE_2D, screenMap2 ) )
		printf ( "buffer2 error with color attachment\n");

	if ( !buffer2.isOk () )
		printf ( "Error with framebuffer 2\n" );
					
	buffer2.unbind ();
	
	unsigned screenMap3 = buffer3.createColorTexture ( GL_RGBA, format, GL_CLAMP, FrameBuffer :: filterLinear );

	buffer3.create ();
	buffer3.bind   ();
	if ( !buffer3.attachColorTexture ( GL_TEXTURE_2D, screenMap3 ) )
		printf ( "buffer3 error with color attachment\n");

	if ( !buffer3.isOk () )
		printf ( "Error with framebuffer 3\n" );
					
	buffer3.unbind ();
	
    glutMainLoop ();

	return 0;
}
